import os.path as op

import pytest

import numpy as np
import numpy.testing as npt

import nibabel as nib
import dipy.data as dpd
import dipy.data.fetcher as fetcher
import dipy.tracking.streamline as dts
import dipy.tracking.utils as dtu
from dipy.stats.analysis import afq_profile
from dipy.io.stateful_tractogram import StatefulTractogram, Space

import AFQ.data as afd
import AFQ.tractography as aft
import AFQ.registration as reg
import AFQ.segmentation as seg
import AFQ.dti as dti
from AFQ.utils.volume import patch_up_roi


dpd.fetch_stanford_hardi()
hardi_dir = op.join(fetcher.dipy_home, "stanford_hardi")
hardi_fdata = op.join(hardi_dir, "HARDI150.nii.gz")
hardi_img = nib.load(hardi_fdata)
hardi_fbval = op.join(hardi_dir, "HARDI150.bval")
hardi_fbvec = op.join(hardi_dir, "HARDI150.bvec")
file_dict = afd.read_stanford_hardi_tractography()
mapping = file_dict['mapping.nii.gz']
streamlines = file_dict['tractography_subsampled.trk']
tg = StatefulTractogram(streamlines, hardi_img, Space.RASMM)
tg.to_vox()
streamlines = tg.streamlines

# streamlines = dts.Streamlines(
#     dtu.transform_tracking_output(
#         streamlines[streamlines._lengths > 10],
#         np.linalg.inv(hardi_img.affine)))


def test_segment():

    templates = afd.read_templates()
    bundles = {'CST_L': {'ROIs': [templates['CST_roi1_L'],
                                  templates['CST_roi2_L']],
                         'rules': [True, True],
                         'prob_map': templates['CST_L_prob_map'],
                         'cross_midline': None},
               'CST_R': {'ROIs': [templates['CST_roi1_R'],
                                  templates['CST_roi1_R']],
                         'rules': [True, True],
                         'prob_map': templates['CST_R_prob_map'],
                         'cross_midline': None}}

    segmentation = seg.Segmentation()
    segmentation.segment(bundles,
                         tg,
                         hardi_fdata,
                         hardi_fbval,
                         hardi_fbvec,
                         mapping=mapping)
    fiber_groups = segmentation.fiber_groups

    # We asked for 2 fiber groups:
    npt.assert_equal(len(fiber_groups), 2)
    # Here's one of them:
    CST_R_sl = fiber_groups['CST_R']
    # Let's make sure there are streamlines in there:
    npt.assert_(len(CST_R_sl) > 0)
    # Calculate the tract profile for a volume of all-ones:
    tract_profile = afq_profile(
        np.ones(nib.load(hardi_fdata).shape[:3]),
        CST_R_sl.streamlines, np.eye(4))
    npt.assert_almost_equal(tract_profile, np.ones(100))

    clean_sl = seg.clean_bundle(CST_R_sl)
    npt.assert_equal(len(clean_sl), len(CST_R_sl))

    # What if you don't have probability maps?
    bundles = {'CST_L': {'ROIs': [templates['CST_roi1_L'],
                                  templates['CST_roi2_L']],
                         'rules': [True, True],
                         'cross_midline': False},
               'CST_R': {'ROIs': [templates['CST_roi1_R'],
                                  templates['CST_roi1_R']],
                         'rules': [True, True],
                         'cross_midline': False}}

    segmentation.segment(bundles,
                         tg,
                         hardi_fdata,
                         hardi_fbval,
                         hardi_fbvec,
                         mapping=mapping)
    fiber_groups = segmentation.fiber_groups

    # This condition should still hold
    npt.assert_equal(len(fiber_groups), 2)
    npt.assert_(len(fiber_groups['CST_R']) > 0)

    # Test with the return_idx kwarg set to True:
    segmentation = seg.Segmentation(return_idx=True)
    segmentation.segment(bundles,
                         tg,
                         hardi_fdata,
                         hardi_fbval,
                         hardi_fbvec,
                         mapping=mapping)
    fiber_groups = segmentation.fiber_groups

    npt.assert_equal(len(fiber_groups), 2)
    npt.assert_(len(fiber_groups['CST_R']['sl']) > 0)
    npt.assert_(len(fiber_groups['CST_R']['idx']) > 0)

    # get bundles for reco method
    bundles = afd.read_hcp_atlas_16_bundles()
    bundle_names = ['whole_brain', 'CST_R', 'CST_L']
    for key in list(bundles):
        if key not in bundle_names:
            bundles.pop(key, None)

    # Try recobundles method
    segmentation = seg.Segmentation(seg_algo='Reco',
                                    progressive=False,
                                    greater_than=10,
                                    rm_small_clusters=1,
                                    rng=np.random.RandomState(seed=8))
    fiber_groups = segmentation.segment(bundles, tg,
                                        hardi_fdata,
                                        hardi_fbval,
                                        hardi_fbvec)

    # This condition should still hold
    npt.assert_equal(len(fiber_groups), 2)
    npt.assert_(len(fiber_groups['CST_R']) > 0)

    # Test with the return_idx kwarg set to True:
    segmentation = seg.Segmentation(seg_algo='Reco',
                                    progressive=False,
                                    greater_than=10,
                                    rm_small_clusters=1,
                                    rng=np.random.RandomState(seed=8),
                                    return_idx=True)

    fiber_groups = segmentation.segment(bundles,
                                        tg,
                                        hardi_fdata,
                                        hardi_fbval,
                                        hardi_fbvec)
    fiber_groups = segmentation.fiber_groups

    npt.assert_equal(len(fiber_groups), 2)
    npt.assert_(len(fiber_groups['CST_R']['sl']) > 0)
    npt.assert_(len(fiber_groups['CST_R']['idx']) > 0)

def test_clean_by_endpoints():
    sl = [np.array([[1, 1, 1],
                    [2, 1, 1],
                    [3, 1, 1],
                    [4, 1, 1]]),
          np.array([[1, 1, 2],
                    [2, 1, 2],
                    [3, 1, 2],
                    [4, 1, 2]]),
          np.array([[1, 1, 1],
                    [2, 1, 1],
                    [3, 1, 1]]),
          np.array([[1, 1, 1],
                    [2, 1, 1]])]

    atlas = np.zeros((20, 20, 20))

    # Targets:
    atlas[1, 1, 1] = 1
    atlas[1, 1, 2] = 2
    atlas[4, 1, 1] = 3
    atlas[4, 1, 2] = 4

    clean_sl = seg.clean_by_endpoints(sl, [1, 2], [3, 4], atlas=atlas)
    npt.assert_equal(list(clean_sl), sl[:2])

    clean_results = list(seg.clean_by_endpoints(sl, [1, 2], [3, 4],
                                                atlas=atlas,
                                                return_idx=True))
    clean_idx = []
    clean_sl = []
    for res in clean_results:
        clean_sl.append(res[0])
        clean_idx.append(res[1])
    npt.assert_equal(list(clean_sl), sl[:2])
    npt.assert_equal(clean_idx, np.array([0, 1]))


    # If tol=1, the third streamline also gets included
    clean_sl = seg.clean_by_endpoints(sl, [1, 2], [3, 4], tol=1, atlas=atlas)
    npt.assert_equal(list(clean_sl), sl[:3])

    # Provide the Nx3 array of indices instead.
    idx_start = np.array(np.where(atlas==1)).T
    idx_end = np.array(np.where(atlas==3)).T

    clean_sl = seg.clean_by_endpoints(sl, idx_start, idx_end, atlas=atlas)
    npt.assert_equal(list(clean_sl), np.array([sl[0]]))

    # Sometimes no requirement for one side:
    clean_sl = seg.clean_by_endpoints(sl, [1], None, atlas=atlas)
    npt.assert_equal(list(clean_sl), [sl[0], sl[2], sl[3]])




import os.path as op
import glob

CLASSIFIERS = ["Development Status :: 3 - Alpha",
               "Environment :: Console",
               "Intended Audience :: Science/Research",
               "License :: OSI Approved :: MIT License",
               "Operating System :: OS Independent",
               "Programming Language :: Python",
               "Topic :: Scientific/Engineering"]

# Description should be a one-liner:
description = "pyAFQ: Automated Fiber Quantification ... in Python"

NAME = "pyAFQ"
MAINTAINER = "Ariel Rokem"
MAINTAINER_EMAIL = "arokem@gmail.com"
DESCRIPTION = description
URL = "http://github.com/yeatmanlab/pyAFQ"
DOWNLOAD_URL = ""
LICENSE = "BSD"
AUTHOR = "pyAFQ developers"
AUTHOR_EMAIL = "arokem@gmail.com"
PLATFORMS = "OS Independent"
SCRIPTS = [op.join('bin', op.split(f)[-1]) for f in glob.glob('bin/*')]
PYTHON_REQUIRES='>=3.6'


# -*- coding: utf-8 -*-# -*- coding: utf-8 -*-
import pandas as pd
import dask.dataframe as ddf
import glob
import os
import os.path as op
from pathlib import PurePath
import json

import numpy as np
import nibabel as nib
from scipy.ndimage.morphology import binary_dilation


import dipy.core.gradients as dpg
from dipy.segment.mask import median_otsu
import dipy.data as dpd
import dipy.tracking.utils as dtu
from dipy.io.streamline import save_tractogram, load_tractogram
from dipy.io.stateful_tractogram import StatefulTractogram, Space
from dipy.stats.analysis import afq_profile

import AFQ.data as afd
from AFQ.dti import _fit as dti_fit
from AFQ.csd import _fit as csd_fit
import AFQ.tractography as aft
import dipy.reconst.dti as dpy_dti
import AFQ.utils.streamlines as aus
import AFQ.segmentation as seg
import AFQ.registration as reg
import AFQ.utils.volume as auv
from AFQ.utils.bin import get_default_args


__all__ = ["AFQ", "make_bundle_dict"]


def do_preprocessing():
    raise NotImplementedError


BUNDLES = ["ATR", "CGC", "CST", "HCC", "IFO", "ILF", "SLF", "ARC", "UNC",
           "FA", "FP"]

DIPY_GH = "https://github.com/nipy/dipy/blob/master/dipy/"


def make_bundle_dict(bundle_names=BUNDLES, seg_algo="afq", resample_to=False):
    """
    Create a bundle dictionary, needed for the segmentation

    Parameters
    ----------
    bundle_names : list, optional
        A list of the bundles to be used in this case. Default: all of them

    resample_to : Nifti1Image, optional
        If set, templates will be resampled to the affine and shape of this
        image.
    """
    if seg_algo == "afq":
        templates = afd.read_templates(resample_to=resample_to)
        callosal_templates = afd.read_callosum_templates(
            resample_to=resample_to)
        # For the arcuate, we need to rename a few of these and duplicate the
        # SLF ROI:
        templates['ARC_roi1_L'] = templates['SLF_roi1_L']
        templates['ARC_roi1_R'] = templates['SLF_roi1_R']
        templates['ARC_roi2_L'] = templates['SLFt_roi2_L']
        templates['ARC_roi2_R'] = templates['SLFt_roi2_R']

        afq_bundles = {}
        # Each bundles gets a digit identifier (to be stored in the tractogram)
        uid = 1
        for name in bundle_names:
            # Consider hard coding since we might have different rules for
            # some tracts
            if name in ["FA", "FP"]:
                afq_bundles[name] = {
                    'ROIs': [templates[name + "_L"],
                             templates[name + "_R"],
                             callosal_templates["Callosum_midsag"]],
                    'rules': [True, True, True],
                    'prob_map': templates[name + "_prob_map"],
                    'cross_midline': True,
                    'uid': uid}
                uid += 1
            # SLF is a special case, because it has an exclusion ROI:
            elif name == "SLF":
                for hemi in ['_R', '_L']:
                    afq_bundles[name + hemi] = {
                        'ROIs': [templates[name + '_roi1' + hemi],
                                 templates[name + '_roi2' + hemi],
                                 templates["SLFt_roi2" + hemi]],
                        'rules': [True, True, False],
                        'prob_map': templates[name + hemi + '_prob_map'],
                        'cross_midline': False,
                        'uid': uid}
                    uid += 1
            else:
                for hemi in ['_R', '_L']:
                    afq_bundles[name + hemi] = {
                        'ROIs': [templates[name + '_roi1' + hemi],
                                 templates[name + '_roi2' + hemi]],
                        'rules': [True, True],
                        'prob_map': templates[name + hemi + '_prob_map'],
                        'cross_midline': False,
                        'uid': uid}

                    uid += 1

    elif seg_algo == "reco":
        afq_bundles = {}
        uid = 1
        bundle_dict = afd.read_hcp_atlas_16_bundles()
        afq_bundles["whole_brain"] = bundle_dict["whole_brain"]
        for name in bundle_names:
            if name in ['CCMid', 'CC_ForcepsMajor', 'CC_ForcepsMinor', 'MCP']:
                afq_bundles[name] = bundle_dict[name]
                afq_bundles[name]['uid'] = uid
                uid += 1
            else:
                for hemi in ["_R", "_L"]:
                    afq_bundles[name + hemi] = bundle_dict[name + hemi]
                    afq_bundles[name + hemi]['uid'] = uid
                    uid += 1
    else:
        raise ValueError("Input: %s is not a valid input`seg_algo`" % seg_algo)

    return afq_bundles


def _b0(row):
    b0_file = _get_fname(row, '_b0.nii.gz')
    if not op.exists(b0_file):
        img = nib.load(row['dwi_file'])
        data = img.get_fdata()
        gtab = row['gtab']
        mean_b0 = np.mean(data[..., ~gtab.b0s_mask], -1)
        mean_b0_img = nib.Nifti1Image(mean_b0, img.affine)
        nib.save(mean_b0_img, b0_file)
        meta = dict(b0_threshold=gtab.b0_threshold,
                    source=row['dwi_file'])
        meta_fname = _get_fname(row, '_b0.json')
        afd.write_json(meta_fname, meta)
    return b0_file


def _brain_mask(row, median_radius=4, numpass=1, autocrop=False,
                vol_idx=None, dilate=10):
    brain_mask_file = _get_fname(row, '_brain_mask.nii.gz')
    if not op.exists(brain_mask_file):
        b0_file = _b0(row)
        mean_b0_img = nib.load(b0_file)
        mean_b0 = mean_b0_img.get_fdata()
        _, brain_mask = median_otsu(mean_b0, median_radius, numpass,
                                    autocrop, dilate=dilate)
        be_img = nib.Nifti1Image(brain_mask.astype(int),
                                 mean_b0_img.affine)
        nib.save(be_img, brain_mask_file)
        meta = dict(source=b0_file,
                    median_radius=median_radius,
                    numpass=numpass,
                    autocrop=autocrop,
                    vol_idx=vol_idx)
        meta_fname = _get_fname(row, '_brain_mask.json')
        afd.write_json(meta_fname, meta)
    return brain_mask_file


def _dti_fit(row):
    dti_params_file = _dti(row)
    dti_params = nib.load(dti_params_file).get_fdata()
    tm = dpy_dti.TensorModel(row['gtab'])
    tf = dpy_dti.TensorFit(tm, dti_params)
    return tf


def _dti(row):
    dti_params_file = _get_fname(row, '_model-DTI_diffmodel.nii.gz')
    if not op.exists(dti_params_file):
        img = nib.load(row['dwi_file'])
        data = img.get_fdata()
        gtab = row['gtab']
        brain_mask_file = _brain_mask(row)
        mask = nib.load(brain_mask_file).get_fdata()
        dtf = dti_fit(gtab, data, mask=mask)
        nib.save(nib.Nifti1Image(dtf.model_params, row['dwi_affine']),
                 dti_params_file)
        meta_fname = _get_fname(row, '_model-DTI_diffmodel.json')
        meta = dict(
            Parameters=dict(
                FitMethod="WLS"),
            OutlierRejection=False,
            ModelURL=f"{DIPY_GH}reconst/dti.py")
        afd.write_json(meta_fname, meta)
    return dti_params_file


def _csd(row, response=None, sh_order=8, lambda_=1, tau=0.1,):
    csd_params_file = _get_fname(row, '_model-CSD_diffmodel.nii.gz')
    if not op.exists(csd_params_file):
        img = nib.load(row['dwi_file'])
        data = img.get_fdata()
        gtab = row['gtab']
        brain_mask_file = _brain_mask(row)
        mask = nib.load(brain_mask_file).get_fdata()
        csdf = csd_fit(gtab, data, mask=mask,
                       response=response, sh_order=sh_order,
                       lambda_=lambda_, tau=tau)
        nib.save(nib.Nifti1Image(csdf.shm_coeff, row['dwi_affine']),
                 csd_params_file)
        meta_fname = _get_fname(row, '_model-CSD_diffmodel.json')
        meta = dict(SphericalHarmonicDegree=sh_order,
                    ResponseFunctionTensor=response,
                    SphericalHarmonicBasis="DESCOTEAUX",
                    ModelURL=f"{DIPY_GH}reconst/csdeconv.py",
                    lambda_=lambda_,
                    tau=tau)
        afd.write_json(meta_fname, meta)
    return csd_params_file


def _dti_fa(row):
    dti_fa_file = _get_fname(row, '_model-DTI_FA.nii.gz')
    if not op.exists(dti_fa_file):
        tf = _dti_fit(row)
        fa = tf.fa
        nib.save(nib.Nifti1Image(fa, row['dwi_affine']),
                 dti_fa_file)
        meta_fname = _get_fname(row, '_model-DTI_FA.json')
        meta = dict()
        afd.write_json(meta_fname, meta)
    return dti_fa_file


def _dti_cfa(row):
    dti_cfa_file = _get_fname(row, '_model-DTI_desc-DEC_FA.nii.gz')
    if not op.exists(dti_cfa_file):
        tf = _dti_fit(row)
        cfa = tf.color_fa
        nib.save(nib.Nifti1Image(cfa, row['dwi_affine']),
                 dti_cfa_file)
        meta_fname = _get_fname(row, '_model-DTI_desc-DEC_FA.json')
        meta = dict()
        afd.write_json(meta_fname, meta)
    return dti_cfa_file


def _dti_pdd(row):
    dti_pdd_file = _get_fname(row, '_model-DTI_PDD.nii.gz')
    if not op.exists(dti_pdd_file):
        tf = _dti_fit(row)
        pdd = tf.directions.squeeze()
        # Invert the x coordinates:
        pdd[..., 0] = pdd[..., 0] * -1

        nib.save(nib.Nifti1Image(pdd, row['dwi_affine']),
                 dti_pdd_file)
        meta_fname = _get_fname(row, '_model-DTI_PDD.json')
        meta = dict()
        afd.write_json(meta_fname, meta)
    return dti_pdd_file


def _dti_md(row):
    dti_md_file = _get_fname(row, '_model-DTI_MD.nii.gz')
    if not op.exists(dti_md_file):
        tf = _dti_fit(row)
        md = tf.md
        nib.save(nib.Nifti1Image(md, row['dwi_affine']),
                 dti_md_file)
        meta_fname = _get_fname(row, '_model-DTI_MD.json')
        meta = dict()
        afd.write_json(meta_fname, meta)
    return dti_md_file


# Keep track of functions that compute scalars:
_scalar_dict = {"dti_fa": _dti_fa,
                "dti_md": _dti_md}


def _reg_prealign(row):
    prealign_file = _get_fname(row, '_prealign_from-DWI_to-MNI_xfm.npy')
    if not op.exists(prealign_file):
        moving = nib.load(_b0(row))
        static = dpd.read_mni_template()
        moving_data = moving.get_fdata()
        moving_affine = moving.affine
        static_data = static.get_fdata()
        static_affine = static.affine
        _, aff = reg.affine_registration(moving_data,
                                         static_data,
                                         moving_affine,
                                         static_affine)
        np.save(prealign_file, aff)
        meta_fname = _get_fname(row, 'prealign_from-DWI_to-MNI_xfm.json')
        meta = dict(type="rigid")
        afd.write_json(meta_fname, meta)
    return prealign_file


def _mapping(row, reg_template):
    mapping_file = _get_fname(row, '_mapping_from-DWI_to_MNI_xfm.nii.gz')
    if not op.exists(mapping_file):
        gtab = row['gtab']
        reg_prealign = np.load(_reg_prealign(row))
        warped_b0, mapping = reg.syn_register_dwi(row['dwi_file'], gtab,
                                                  template=reg_template,
                                                  prealign=reg_prealign)
        mapping.codomain_world2grid = np.linalg.inv(reg_prealign)
        reg.write_mapping(mapping, mapping_file)
        meta_fname = _get_fname(row, '_mapping_reg_prealign.json')
        meta = dict(type="displacementfield")
        afd.write_json(meta_fname, meta)

    return mapping_file


def _wm_mask(row, wm_labels, wm_fa_thresh=0.2):
    wm_mask_file = _get_fname(row, '_wm_mask.nii.gz')
    if not op.exists(wm_mask_file):
        dwi_img = nib.load(row['dwi_file'])
        dwi_data = dwi_img.get_fdata()

        if 'seg_file' in row.index:
            # If we found a white matter segmentation in the
            # expected location:
            seg_img = nib.load(row['seg_file'])
            seg_data_orig = seg_img.get_fdata()
            # For different sets of labels, extract all the voxels that
            # have any of these values:
            wm_mask = np.sum(np.concatenate([(seg_data_orig == l)[..., None]
                                            for l in wm_labels], -1), -1)

            # Resample to DWI data:
            wm_mask = np.round(reg.resample(wm_mask, dwi_data[..., 0],
                                            seg_img.affine,
                                            dwi_img.affine)).astype(int)
            meta = dict(source=row['seg_file'],
                        wm_labels=wm_labels)
        else:
            # Otherwise, we'll identify the white matter based on FA:
            fa_fname = _dti_fa(row)
            dti_fa = nib.load(fa_fname).get_fdata()
            wm_mask = dti_fa > wm_fa_thresh
            meta = dict(source=fa_fname,
                        fa_threshold=wm_fa_thresh)

        # Dilate to be sure to reach the gray matter:
        wm_mask = binary_dilation(wm_mask) > 0

        nib.save(nib.Nifti1Image(wm_mask.astype(int), row['dwi_affine']),
                 wm_mask_file)

        meta_fname = _get_fname(row, '_wm_mask.json')
        afd.write_json(meta_fname, meta)

    return wm_mask_file


def _streamlines(row, wm_labels, tracking_params=None):
    """
    wm_labels : list
        The values within the segmentation that are considered white matter. We
        will use this part of the image both to seed tracking (seeding
        throughout), and for stopping.
    """
    if tracking_params is None:
        tracking_params = get_default_args(aft.track)

    odf_model = tracking_params["odf_model"]
    directions = tracking_params["directions"]

    streamlines_file = _get_fname(
        row,
        f'_space-RASMM_model-{odf_model}_desc-{directions}_tractography.trk')

    if not op.exists(streamlines_file):
        if odf_model == "DTI":
            params_file = _dti(row)
        elif odf_model == "CSD":
            params_file = _csd(row)
        wm_mask_fname = _wm_mask(row, wm_labels)
        wm_mask = nib.load(wm_mask_fname).get_fdata().astype(bool)
        tracking_params['seed_mask'] = wm_mask
        tracking_params['stop_mask'] = wm_mask
        sft = aft.track(params_file, **tracking_params)
        sft.to_vox()
        meta_directions = {"det": "deterministic",
                           "prob": "probabilistic"}

        meta = dict(
            TractographyClass="local",
            TractographyMethod=meta_directions[tracking_params["directions"]],
            Count=len(sft.streamlines),
            Seeding=dict(
                ROI=wm_mask_fname,
                n_seeds=tracking_params["n_seeds"],
                random_seeds=tracking_params["random_seeds"]),
            Constraints=dict(AnatomicalImage=wm_mask_fname),
            Parameters=dict(Units="mm",
                            StepSize=tracking_params["step_size"],
                            MinimumLength=tracking_params["min_length"],
                            MaximumLength=tracking_params["max_length"],
                            Unidirectional=False))

        meta_fname = _get_fname(
            row,
            f'_space-RASMM_model-{odf_model}_desc-'
            f'{directions}_tractography.json')
        afd.write_json(meta_fname, meta)
        save_tractogram(sft, streamlines_file, bbox_valid_check=False)

    return streamlines_file


def _segment(row, wm_labels, bundle_dict, reg_template,
             tracking_params, segmentation_params, clean_params):
    # We pass `clean_params` here, but do not use it, so we have the
    # same signature as `_clean_bundles`.
    odf_model = tracking_params["odf_model"]
    directions = tracking_params["directions"]
    seg_algo = segmentation_params["seg_algo"]
    bundles_file = _get_fname(
        row,
        f'_space-RASMM_model-{odf_model}_desc-{directions}-'
        f'{seg_algo}_tractography.trk')

    if not op.exists(bundles_file):
        streamlines_file = _streamlines(
            row,
            wm_labels,
            tracking_params)

        img = nib.load(row['dwi_file'])
        tg = load_tractogram(streamlines_file, img, Space.VOX)
        reg_prealign = np.load(_reg_prealign(row))

        segmentation = seg.Segmentation(**segmentation_params)
        bundles = segmentation.segment(bundle_dict,
                                       tg,
                                       row['dwi_file'],
                                       row['bval_file'],
                                       row['bvec_file'],
                                       reg_template=reg_template,
                                       mapping=_mapping(row, reg_template),
                                       reg_prealign=reg_prealign)

        if segmentation_params['return_idx']:
            idx = {bundle: bundles[bundle]['idx'].tolist()
                   for bundle in bundle_dict}
            afd.write_json(bundles_file.split('.')[0] + '_idx.json',
                           idx)
            bundles = {bundle: bundles[bundle]['sl'] for bundle in bundle_dict}

        tgram = aus.bundles_to_tgram(bundles, bundle_dict, img)
        save_tractogram(tgram, bundles_file)
        meta = dict(source=streamlines_file,
                    Parameters=segmentation_params)
        meta_fname = bundles_file.split('.')[0] + '.json'
        afd.write_json(meta_fname, meta)

    return bundles_file


def _clean_bundles(row, wm_labels, bundle_dict, reg_template, tracking_params,
                   segmentation_params, clean_params):
    odf_model = tracking_params['odf_model']
    directions = tracking_params['directions']
    seg_algo = segmentation_params['seg_algo']
    clean_bundles_file = _get_fname(
        row,
        f'_space-RASMM_model-{odf_model}_desc-{directions}-'
        f'{seg_algo}-clean_tractography.trk')

    if not op.exists(clean_bundles_file):
        bundles_file = _segment(row,
                                wm_labels,
                                bundle_dict,
                                reg_template,
                                tracking_params,
                                segmentation_params,
                                clean_params)

        sft = load_tractogram(bundles_file,
                              row['dwi_img'],
                              Space.VOX)

        tgram = nib.streamlines.Tractogram([], {'bundle': []})
        if clean_params['return_idx']:
            return_idx = {}

        for b in bundle_dict.keys():
            if b != "whole_brain":
                idx = np.where(sft.data_per_streamline['bundle']
                               == bundle_dict[b]['uid'])[0]
                this_tg = StatefulTractogram(
                    sft.streamlines[idx],
                    row['dwi_img'],
                    Space.VOX)
                this_tg = seg.clean_bundle(this_tg, **clean_params)
                if clean_params['return_idx']:
                    this_tg, this_idx = this_tg
                    idx_file = bundles_file.split('.')[0] + '_idx.json'
                    with open(idx_file) as ff:
                        bundle_idx = json.load(ff)[b]
                    return_idx[b] = np.array(bundle_idx)[this_idx].tolist()
                this_tgram = nib.streamlines.Tractogram(
                    this_tg.streamlines,
                    data_per_streamline={
                        'bundle': (len(this_tg)
                                   * [bundle_dict[b]['uid']])},
                        affine_to_rasmm=row['dwi_affine'])
                tgram = aus.add_bundles(tgram, this_tgram)
        save_tractogram(
            StatefulTractogram(tgram.streamlines,
                               sft,
                               Space.VOX,
                               data_per_streamline=tgram.data_per_streamline),
            clean_bundles_file)

        seg_args = get_default_args(seg.clean_bundle)
        for k in seg_args:
            if callable(seg_args[k]):
                seg_args[k] = seg_args[k].__name__

        meta = dict(source=bundles_file,
                    Parameters=seg_args)
        meta_fname = clean_bundles_file.split('.')[0] + '.json'
        afd.write_json(meta_fname, meta)

        if clean_params['return_idx']:
            afd.write_json(clean_bundles_file.split('.')[0] + '_idx.json',
                           return_idx)

    return clean_bundles_file


def _tract_profiles(row, wm_labels, bundle_dict, reg_template,
                    tracking_params, segmentation_params, clean_params,
                    scalars, weighting=None):
    profiles_file = _get_fname(row, '_profiles.csv')
    if not op.exists(profiles_file):
        bundles_file = _clean_bundles(row,
                                      wm_labels,
                                      bundle_dict,
                                      reg_template,
                                      tracking_params,
                                      segmentation_params,
                                      clean_params)
        keys = []
        vals = []
        for k in bundle_dict.keys():
            if k != "whole_brain":
                keys.append(bundle_dict[k]['uid'])
                vals.append(k)
        reverse_dict = dict(zip(keys, vals))

        bundle_names = []
        profiles = []
        node_numbers = []
        scalar_names = []

        trk = nib.streamlines.load(bundles_file)
        for scalar in scalars:
            scalar_file = _scalar_dict[scalar](row)
            scalar_data = nib.load(scalar_file).get_fdata()
            for b in np.unique(trk.tractogram.data_per_streamline['bundle']):
                idx = np.where(
                    trk.tractogram.data_per_streamline['bundle'] == b)[0]
                this_sl = trk.streamlines[idx]
                bundle_name = reverse_dict[b]
                this_profile = afq_profile(
                    scalar_data,
                    this_sl,
                    row["dwi_affine"])
                nodes = list(np.arange(this_profile.shape[0]))
                bundle_names.extend([bundle_name] * len(nodes))
                node_numbers.extend(nodes)
                scalar_names.extend([scalar] * len(nodes))
                profiles.extend(list(this_profile))

        profile_dframe = pd.DataFrame(dict(profiles=profiles,
                                           bundle=bundle_names,
                                           node=node_numbers,
                                           scalar=scalar_names))
        profile_dframe.to_csv(profiles_file)
        meta = dict(source=bundles_file,
                    parameters=get_default_args(afq_profile))
        meta_fname = profiles_file.split('.')[0] + '.json'
        afd.write_json(meta_fname, meta)

    return profiles_file


def _template_xform(row, reg_template):
    template_xform_file = _get_fname(row, "_template_xform.nii.gz")
    if not op.exists(template_xform_file):
        reg_prealign = np.load(_reg_prealign(row))
        mapping = reg.read_mapping(_mapping(row,
                                            reg_template),
                                   row['dwi_file'],
                                   reg_template,
                                   prealign=np.linalg.inv(reg_prealign))

        template_xform = mapping.transform_inverse(reg_template.get_fdata())
        nib.save(nib.Nifti1Image(template_xform,
                                 row['dwi_affine']),
                 template_xform_file)

    return template_xform_file


def _export_rois(row, bundle_dict, reg_template):
    reg_prealign = np.load(_reg_prealign(row))

    mapping = reg.read_mapping(_mapping(row, reg_template),
                               row['dwi_file'],
                               reg_template,
                               prealign=np.linalg.inv(reg_prealign))

    rois_dir = op.join(row['results_dir'], 'ROIs')
    os.makedirs(rois_dir, exist_ok=True)

    for bundle in bundle_dict:
        for ii, roi in enumerate(bundle_dict[bundle]['ROIs']):

            if bundle_dict[bundle]['rules'][ii]:
                inclusion = 'include'
            else:
                inclusion = 'exclude'

            warped_roi = auv.patch_up_roi(
                (mapping.transform_inverse(
                    roi.get_fdata(),
                    interpolation='linear')) > 0).astype(int)

            fname = op.split(
                _get_fname(
                    row,
                    f'_desc-ROI-{bundle}-{ii + 1}-{inclusion}.nii.gz'))

            fname = op.join(fname[0], rois_dir, fname[1])

            # Cast to float32, so that it can be read in by MI-Brain:
            nib.save(nib.Nifti1Image(warped_roi.astype(np.float32),
                                     row['dwi_affine']),
                     fname)
            meta = dict()
            meta_fname = fname.split('.')[0] + '.json'
            afd.write_json(meta_fname, meta)


def _export_bundles(row, wm_labels, bundle_dict, reg_template,
                    tracking_params, segmentation_params, clean_params):

    odf_model = tracking_params['odf_model']
    directions = tracking_params['directions']
    seg_algo = segmentation_params['seg_algo']

    for func, folder in zip([_clean_bundles, _segment],
                            ['clean_bundles', 'bundles']):
        bundles_file = func(row,
                            wm_labels,
                            bundle_dict,
                            reg_template,
                            tracking_params,
                            segmentation_params,
                            clean_params)

        bundles_dir = op.join(row['results_dir'], folder)
        os.makedirs(bundles_dir, exist_ok=True)
        trk = nib.streamlines.load(bundles_file)
        tg = trk.tractogram
        streamlines = tg.streamlines
        for bundle in bundle_dict:
            if bundle != "whole_brain":
                uid = bundle_dict[bundle]['uid']
                idx = np.where(tg.data_per_streamline['bundle'] == uid)[0]
                this_sl = dtu.transform_tracking_output(
                    streamlines[idx],
                    np.linalg.inv(row['dwi_affine']))

                this_tgm = StatefulTractogram(this_sl, row['dwi_img'],
                                              Space.VOX)

                fname = op.split(
                    _get_fname(
                        row,
                        f'_space-RASMM_model-{odf_model}_desc-{directions}-'
                        f'{seg_algo}-{bundle}_tractography.trk'))
                fname = op.join(fname[0], bundles_dir, fname[1])
                save_tractogram(this_tgm, fname, bbox_valid_check=False)
                meta = dict(source=bundles_file)
                meta_fname = fname.split('.')[0] + '.json'
                afd.write_json(meta_fname, meta)


def _get_affine(fname):
    return nib.load(fname).affine


def _get_fname(row, suffix):
    split_fdwi = op.split(row['dwi_file'])
    fname = op.join(row['results_dir'], split_fdwi[1].split('.')[0]
                    + suffix)
    return fname


class AFQ(object):
    """

    This is file folder structure that AFQ requires in your study folder::

|    study
|      ├-derivatives
|            ├-dmriprep
|                ├── sub01
|                │   ├── sess01
|                │   │   ├── anat
|                │   │   │   ├── sub-01_sess-01_aparc+aseg.nii.gz
|                │   │   │   └── sub-01_sess-01_T1.nii.gz
|                │   │   └── dwi
|                │   │       ├── sub-01_sess-01_dwi.bvals
|                │   │       ├── sub-01_sess-01_dwi.bvecs
|                │   │       └── sub-01_sess-01_dwi.nii.gz
|                │   └── sess02
|                │       ├── anat
|                │       │   ├── sub-01_sess-02_aparc+aseg.nii.gz
|                │       │   └── sub-01_sess-02_T1w.nii.gz
|                │       └── dwi
|                │           ├── sub-01_sess-02_dwi.bvals
|                │           ├── sub-01_sess-02_dwi.bvecs
|                │           └── sub-01_sess-02_dwi.nii.gz
|                └── sub02
|                   ├── sess01
|                   │   ├── anat
|                   │       ├── sub-02_sess-01_aparc+aseg.nii.gz
|                   │   │   └── sub-02_sess-01_T1w.nii.gz
|                   │   └── dwi
|                   │       ├── sub-02_sess-01_dwi.bvals
|                   │       ├── sub-02_sess-01_dwi.bvecs
|                   │       └── sub-02_sess-01_dwi.nii.gz
|                   └── sess02
|                       ├── anat
|                       │   ├── sub-02_sess-02_aparc+aseg.nii.gz
|                       │   └── sub-02_sess-02_T1w.nii.gz
|                       └── dwi
|                           ├── sub-02_sess-02_dwi.bvals
|                           ├── sub-02_sess-02_dwi.bvecs
|                           └── sub-02_sess-02_dwi.nii.gz

    This structure can be automatically generated from BIDS-compliant
    data [1]_, using the dmriprep software [2]_ and BIDS app.

    Notes
    -----
    The structure of the file-system required here resembles that specified
    by BIDS [1]_. In the future, this will be organized according to the
    BIDS derivatives specification, as we require preprocessed, rather than
    raw data.

    .. [1] Gorgolewski et al. (2016). The brain imaging data structure,
           a format for organizing and describing outputs of neuroimaging
           experiments. Scientific Data, 3::160044. DOI: 10.1038/sdata.2016.44.

    .. [2] https://github.com/nipy/dmriprep

    """
    def __init__(self,
                 dmriprep_path,
                 sub_prefix="sub",
                 dwi_folder="dwi",
                 dwi_file="*dwi",
                 anat_folder="anat",
                 anat_file="*T1w*",
                 seg_file='*aparc+aseg*',
                 b0_threshold=0,
                 bundle_names=BUNDLES,
                 dask_it=False,
                 reg_template=None,
                 scalars=["dti_fa", "dti_md"],
                 wm_labels=[250, 251, 252, 253, 254, 255, 41, 2, 16, 77],
                 tracking_params=None,
                 segmentation_params=None,
                 clean_params=None):
        """

        dmriprep_path: str
            The path to the preprocessed diffusion data.

        seg_algo : str
            Which algorithm to use for segmentation.
            Can be one of: {"afq", "reco"}

        b0_threshold : int, optional
            The value of b under which it is considered to be b0. Default: 0.

        odf_model : string, optional
            Which model to use for determining directions in tractography
            {"DTI", "DKI", "CSD"}. Default: "DTI"

        directions : string, optional
            How to select directions for tracking (deterministic or
            probablistic) {"det", "prob"}. Default: "det".

        dask_it : bool, optional
            Whether to use a dask DataFrame object

        wm_labels : list, optional
            A list of the labels of the white matter in the segmentation file
            used. Default: the white matter values for the segmentation
            provided with the HCP data, including labels for midbrain:
            [250, 251, 252, 253, 254, 255, 41, 2, 16, 77].

        segmentation_params : dict, optional
            The parameters for segmentation. Default: use the default behavior
            of the seg.Segmentation object

        tracking_params: dict, optional
            The parameters for tracking. Default: use the default behavior of
            the aft.track function.
        """

        self.wm_labels = wm_labels

        self.scalars = scalars

        default_tracking_params = get_default_args(aft.track)
        # Replace the defaults only for kwargs for which a non-default value was
        # given:
        if tracking_params is not None:
            for k in tracking_params:
                default_tracking_params[k] = tracking_params[k]

        self.tracking_params = default_tracking_params

        default_seg_params = get_default_args(seg.Segmentation.__init__)
        if segmentation_params is not None:
            for k in segmentation_params:
                default_seg_params[k] = segmentation_params[k]

        self.segmentation_params = default_seg_params
        self.seg_algo = self.segmentation_params["seg_algo"].lower()
        self.bundle_dict = make_bundle_dict(bundle_names=bundle_names,
                                            seg_algo=self.seg_algo,
                                            resample_to=reg_template)

        default_clean_params = get_default_args(seg.clean_bundle)
        if clean_params is not None:
            for k in clean_params:
                default_clean_params[k] = clean_params[k]

        self.clean_params = default_clean_params

        if reg_template is None:
            self.reg_template = dpd.read_mni_template()
        else:
            if not isinstance(reg_template, nib.Nifti1Image):
                reg_template = nib.load(reg_template)
            self.reg_template = reg_template
        # This is the place in which each subject's full data lives
        self.dmriprep_dirs = glob.glob(op.join(dmriprep_path,
                                               '%s*' % sub_prefix))

        # This is where all the outputs will go:
        self.afq_dir = op.join(
            op.join(*PurePath(dmriprep_path).parts[:-1]), 'afq')

        os.makedirs(self.afq_dir, exist_ok=True)

        self.subjects = [op.split(p)[-1] for p in self.dmriprep_dirs]

        sub_list = []
        sess_list = []
        dwi_file_list = []
        bvec_file_list = []
        bval_file_list = []
        anat_file_list = []
        seg_file_list = []
        results_dir_list = []
        for subject, sub_dir in zip(self.subjects, self.dmriprep_dirs):
            sessions = glob.glob(op.join(sub_dir, '*'))
            for sess in sessions:
                results_dir_list.append(op.join(self.afq_dir,
                                        subject,
                                        PurePath(sess).parts[-1]))

                os.makedirs(results_dir_list[-1], exist_ok=True)

                dwi_file_list.append(glob.glob(op.join(sub_dir,
                                                       ('%s/%s/%s.nii.gz' %
                                                        (sess, dwi_folder,
                                                         dwi_file))))[0])

                bvec_file_list.append(glob.glob(op.join(sub_dir,
                                                        ('%s/%s/%s.bvec*' %
                                                         (sess, dwi_folder,
                                                          dwi_file))))[0])

                bval_file_list.append(glob.glob(op.join(sub_dir,
                                                        ('%s/%s/%s.bval*' %
                                                         (sess, dwi_folder,
                                                          dwi_file))))[0])

                # The following two may or may not exist:
                this_anat_file = glob.glob(op.join(sub_dir,
                                           ('%s/%s/%s.nii.gz' %
                                            (sess,
                                             anat_folder,
                                             anat_file))))
                if len(this_anat_file):
                    anat_file_list.append(this_anat_file[0])

                this_seg_file = glob.glob(op.join(sub_dir,
                                                  ('%s/%s/%s.nii.gz' %
                                                   (sess,
                                                    anat_folder,
                                                    seg_file))))
                if len(this_seg_file):
                    seg_file_list.append(this_seg_file[0])

                sub_list.append(subject)
                sess_list.append(sess)
        self.data_frame = pd.DataFrame(dict(subject=sub_list,
                                            dwi_file=dwi_file_list,
                                            bvec_file=bvec_file_list,
                                            bval_file=bval_file_list,
                                            sess=sess_list,
                                            results_dir=results_dir_list))
        # Add these if they exist:
        if len(seg_file_list):
            self.data_frame['seg_file'] = seg_file_list
        if len(anat_file_list):
            self.data_frame['anat_file'] = anat_file_list

        if dask_it:
            self.data_frame = ddf.from_pandas(self.data_frame,
                                              npartitions=len(sub_list))
        self.set_gtab(b0_threshold)
        self.set_dwi_affine()
        self.set_dwi_img()

    def set_gtab(self, b0_threshold):
        self.data_frame['gtab'] = self.data_frame.apply(
            lambda x: dpg.gradient_table(x['bval_file'], x['bvec_file'],
                                         b0_threshold=b0_threshold),
            axis=1)

    def get_gtab(self):
        return self.data_frame['gtab']

    gtab = property(get_gtab, set_gtab)

    def set_dwi_affine(self):
        self.data_frame['dwi_affine'] =\
            self.data_frame['dwi_file'].apply(_get_affine)

    def get_dwi_affine(self):
        return self.data_frame['dwi_affine']

    dwi_affine = property(get_dwi_affine, set_dwi_affine)

    def set_dwi_img(self):
        self.data_frame['dwi_img'] =\
            self.data_frame['dwi_file'].apply(nib.load)

    def get_dwi_img(self):
        return self.data_frame['dwi_img']

    dwi_img = property(get_dwi_img, set_dwi_img)

    def __getitem__(self, k):
        return self.data_frame.__getitem__(k)

    def set_b0(self):
        if 'b0_file' not in self.data_frame.columns:
            self.data_frame['b0_file'] =\
                self.data_frame.apply(_b0,
                                      axis=1)

    def get_b0(self):
        self.set_b0()
        return self.data_frame['b0_file']

    b0 = property(get_b0, set_b0)

    def set_brain_mask(self):
        if 'brain_mask_file' not in self.data_frame.columns:
            self.data_frame['brain_mask_file'] =\
                self.data_frame.apply(_brain_mask,
                                      axis=1)

    def get_brain_mask(self):
        self.set_brain_mask()
        return self.data_frame['brain_mask_file']

    brain_mask = property(get_brain_mask, set_brain_mask)

    def set_wm_mask(self):
        if 'wm_mask_file' not in self.data_frame.columns:
            self.data_frame['wm_mask_file'] =\
                self.data_frame.apply(_wm_mask,
                                      args=[self.wm_labels],
                                      axis=1)

    def get_wm_mask(self):
        self.set_wm_mask()
        return self.data_frame['wm_mask_file']

    wm_mask = property(get_wm_mask, set_wm_mask)

    def set_dti(self):
        if 'dti_params_file' not in self.data_frame.columns:
            self.data_frame['dti_params_file'] =\
                self.data_frame.apply(_dti,
                                      axis=1)

    def get_dti(self):
        self.set_dti()
        return self.data_frame['dti_params_file']

    dti = property(get_dti, set_dti)

    def set_dti_fa(self):
        if 'dti_fa_file' not in self.data_frame.columns:
            self.data_frame['dti_fa_file'] =\
                self.data_frame.apply(_dti_fa,
                                      axis=1)

    def get_dti_fa(self):
        self.set_dti_fa()
        return self.data_frame['dti_fa_file']

    dti_fa = property(get_dti_fa, set_dti_fa)

    def set_dti_cfa(self):
        if 'dti_cfa_file' not in self.data_frame.columns:
            self.data_frame['dti_cfa_file'] =\
                self.data_frame.apply(_dti_cfa,
                                      axis=1)

    def get_dti_cfa(self):
        self.set_dti_cfa()
        return self.data_frame['dti_cfa_file']

    dti_cfa = property(get_dti_cfa, set_dti_cfa)

    def set_dti_pdd(self):
        if 'dti_pdd_file' not in self.data_frame.columns:
            self.data_frame['dti_pdd_file'] =\
                self.data_frame.apply(_dti_pdd,
                                      axis=1)

    def get_dti_pdd(self):
        self.set_dti_pdd()
        return self.data_frame['dti_pdd_file']

    dti_pdd = property(get_dti_pdd, set_dti_pdd)

    def set_dti_md(self):
        if 'dti_md_file' not in self.data_frame.columns:
            self.data_frame['dti_md_file'] =\
                self.data_frame.apply(_dti_md,
                                      axis=1)

    def get_dti_md(self):
        self.set_dti_md()
        return self.data_frame['dti_md_file']

    dti_md = property(get_dti_md, set_dti_md)

    def set_mapping(self):
        if 'mapping' not in self.data_frame.columns:
            self.data_frame['mapping'] =\
                self.data_frame.apply(_mapping,
                                      args=[self.reg_template],
                                      axis=1)

    def get_mapping(self):
        self.set_mapping()
        return self.data_frame['mapping']

    mapping = property(get_mapping, set_mapping)

    def set_streamlines(self):
        if 'streamlines_file' not in self.data_frame.columns:
            self.data_frame['streamlines_file'] =\
                self.data_frame.apply(_streamlines, axis=1,
                                      args=[self.wm_labels,
                                            self.tracking_params])

    def get_streamlines(self):
        self.set_streamlines()
        return self.data_frame['streamlines_file']

    streamlines = property(get_streamlines, set_streamlines)

    def set_bundles(self):
        if 'bundles_file' not in self.data_frame.columns:
            self.data_frame['bundles_file'] =\
                self.data_frame.apply(
                    _segment,
                    axis=1,
                    args=[self.wm_labels,
                          self.bundle_dict,
                          self.reg_template,
                          self.tracking_params,
                          self.segmentation_params,
                          self.clean_params])

    def get_bundles(self):
        self.set_bundles()
        return self.data_frame['bundles_file']

    bundles = property(get_bundles, set_bundles)

    def set_clean_bundles(self):
        if 'clean_bundles_file' not in self.data_frame.columns:
            if self.seg_algo == "reco":
                self.data_frame['clean_bundles_file'] =\
                    self.data_frame['bundles_file']
            else:
                self.data_frame['clean_bundles_file'] =\
                    self.data_frame.apply(_clean_bundles, axis=1,
                                          args=[self.wm_labels,
                                                self.bundle_dict,
                                                self.reg_template,
                                                self.tracking_params,
                                                self.segmentation_params,
                                                self.clean_params])

    def get_clean_bundles(self):
        self.set_clean_bundles()
        return self.data_frame['clean_bundles_file']

    clean_bundles = property(get_clean_bundles, set_clean_bundles)

    def set_tract_profiles(self):
        if 'tract_profiles_file' not in self.data_frame.columns:
            self.data_frame['tract_profiles_file'] =\
                self.data_frame.apply(_tract_profiles,
                                      args=[self.wm_labels,
                                            self.bundle_dict,
                                            self.reg_template,
                                            self.tracking_params,
                                            self.segmentation_params,
                                            self.clean_params,
                                            self.scalars],
                                      axis=1)

    def get_tract_profiles(self):
        self.set_tract_profiles()
        return self.data_frame['tract_profiles_file']

    tract_profiles = property(get_tract_profiles, set_tract_profiles)

    def set_template_xform(self):
        if 'template_xform_file' not in self.data_frame.columns:
            self.data_frame['template_xform_file'] = \
                self.data_frame.apply(_template_xform,
                                      args=[self.reg_template],
                                      axis=1)

    def get_template_xform(self):
        self.set_template_xform()
        return self.data_frame['template_xform_file']

    template_xform = property(get_template_xform, set_template_xform)

    def export_rois(self):
        self.data_frame.apply(_export_rois,
                              args=[self.bundle_dict,
                                    self.reg_template],
                              axis=1)

    def export_bundles(self):
        self.data_frame.apply(_export_bundles,
                              args=[self.wm_labels,
                                    self.bundle_dict,
                                    self.reg_template,
                                    self.tracking_params,
                                    self.segmentation_params,
                                    self.clean_params],
                              axis=1)

    def combine_profiles(self):
        dfs = []
        for ii, fname in enumerate(self.tract_profiles):
            profiles = pd.read_csv(fname)
            profiles['sub'] = self.data_frame['subject'].iloc[ii]
            profiles['sess'] = op.split(self.data_frame['sess'].iloc[ii])[-1]
            dfs.append(profiles)

        df = pd.concat(dfs)
        df.to_csv(op.join(self.afq_dir, 'tract_profiles.csv'), index=False)
        return df



import os
import os.path as op
import shutil

import numpy as np
import numpy.testing as npt

import nibabel as nib
import nibabel.tmpdirs as nbtmp

from dipy.core.geometry import vector_norm
import dipy.core.gradients as dpg
import dipy.data as dpd

from AFQ import dti
from AFQ._fixes import in_place_norm
from AFQ.utils.testing import make_dti_data


def test_fit_dti():
    # Let's see whether we can pass a list of files for each one:
    fdata1, fbval1, fbvec1 = dpd.get_fnames('small_101D')
    fdata2, fbval2, fbvec2 = dpd.get_fnames('small_101D')

    with nbtmp.InTemporaryDirectory() as tmpdir:
        file_dict = dti.fit_dti([fdata1, fdata2],
                                [fbval1, fbval2],
                                [fbvec1, fbvec2],
                                out_dir=tmpdir)
        for f in file_dict.values():
            npt.assert_(op.exists(f))


def test_predict_dti():
    with nbtmp.InTemporaryDirectory() as tmpdir:
        fbval = op.join(tmpdir, 'dti.bval')
        fbvec = op.join(tmpdir, 'dti.bvec')
        fdata = op.join(tmpdir, 'dti.nii.gz')
        make_dti_data(fbval, fbvec, fdata)
        file_dict = dti.fit_dti(fdata, fbval, fbvec, out_dir=tmpdir)
        params_file = file_dict['params']
        gtab = dpg.gradient_table(fbval, fbvec)
        predict_fname = dti.predict(params_file, gtab, S0_file=fdata,
                                    out_dir=tmpdir)
        prediction = nib.load(predict_fname).get_fdata()
        npt.assert_almost_equal(prediction, nib.load(fdata).get_fdata())

        # If you have a mask into the volume, you will predict only that
        # part of the volume:
        mask = np.zeros(prediction.shape[:3], dtype=bool)
        mask[2:4, 2:4, 2:4] = 1
        file_dict = dti.fit_dti(fdata, fbval, fbvec, mask=mask,
                                out_dir=tmpdir)
        params_file = file_dict['params']
        predict_fname = dti.predict(params_file, gtab, S0_file=fdata,
                                    out_dir=tmpdir)
        prediction = nib.load(predict_fname).get_fdata()
        npt.assert_almost_equal(prediction[mask],
                                nib.load(fdata).get_fdata()[mask])


def test_cli():
    with nbtmp.InTemporaryDirectory() as tmpdir:
        fbval = op.join(tmpdir, 'dti.bval')
        fbvec = op.join(tmpdir, 'dti.bvec')
        fdata = op.join(tmpdir, 'dti.nii.gz')
        make_dti_data(fbval, fbvec, fdata)
        cmd = " ".join(
            ["pyAFQ_dti",
             "-d", op.join(tmpdir, "dti.nii.gz"),
             "-l", op.join(tmpdir, "dti.bval"),
             "-c", op.join(tmpdir, "dti.bvec")])
        out = os.system(cmd)
        assert out ==  0
        assert op.exists(op.join(tmpdir, 'dti', 'dti_params.nii.gz'))


def test_inplace_norm():
    vec = [[8, 15, 0], [0, 36, 77]]
    norm1 = vector_norm(vec)
    norm2 = in_place_norm(vec)
    npt.assert_equal(norm1, norm2)

    vec = [[8.0, 15.0, 0.0], [0.0, 36.0, 77.0]]
    norm1 = vector_norm(vec)
    norm2 = in_place_norm(vec)
    npt.assert_equal(norm1, norm2)

    vec = [[8, 15, 0], [0, 36, 77]]
    norm1 = vector_norm(vec, keepdims=True)
    norm2 = in_place_norm(vec, keepdims=True)
    npt.assert_equal(norm1, norm2)

    vec = [[8, 15, 0], [0, 36, 77]]
    norm1 = vector_norm(vec, axis=0)
    norm2 = in_place_norm(vec, axis=0)
    npt.assert_equal(norm1, norm2)



import tempfile
import os
import os.path as op
import shutil

import toml

import numpy as np
import numpy.testing as npt

import pandas as pd
from pandas.testing import assert_frame_equal

import nibabel as nib
import nibabel.tmpdirs as nbtmp

import dipy.tracking.utils as dtu
import dipy.tracking.streamline as dts
import dipy.data as dpd
from dipy.data import fetcher
from dipy.io.streamline import save_tractogram, load_tractogram
from dipy.io.stateful_tractogram import StatefulTractogram, Space

from AFQ import api
import AFQ.data as afd
import AFQ.segmentation as seg
import AFQ.utils.streamlines as aus


def touch(fname, times=None):
    with open(fname, 'a'):
        os.utime(fname, times)


def create_dummy_dmriprep_path(n_subjects, n_sessions):
    preproc_dir = tempfile.mkdtemp()
    subjects = ['sub-0%s' % (d + 1) for d in range(n_subjects)]
    sessions = ['sess-0%s' % (d + 1) for d in range(n_sessions)]
    for subject in subjects:
        for session in sessions:
            for modality in ['anat', 'dwi']:
                os.makedirs(op.join(preproc_dir, subject, session, modality))
            # Make some dummy data:
            aff = np.eye(4)
            data = np.ones((10, 10, 10, 6))
            bvecs = np.vstack([np.eye(3), np.eye(3)])
            bvecs[0] = 0
            bvals = np.ones(6) * 1000.
            bvals[0] = 0
            np.savetxt(op.join(preproc_dir, subject, session, 'dwi',
                               'dwi.bvals'),
                       bvals)
            np.savetxt(op.join(preproc_dir, subject, session, 'dwi',
                               'dwi.bvecs'),
                       bvecs)
            nib.save(nib.Nifti1Image(data, aff),
                     op.join(preproc_dir, subject, session, 'dwi',
                             'dwi.nii.gz'))
            nib.save(nib.Nifti1Image(data, aff),
                     op.join(preproc_dir, subject, session, 'anat',
                             'T1w.nii.gz'))
            nib.save(nib.Nifti1Image(data, aff),
                     op.join(preproc_dir, subject, session, 'anat',
                             'aparc+aseg.nii.gz'))

    return preproc_dir


def test_AFQ_init():
    """
    Test the initialization of the AFQ object
    """
    n_subjects = 3
    n_sessions = 2
    dmriprep_path = create_dummy_dmriprep_path(n_subjects, n_sessions)
    my_afq = api.AFQ(dmriprep_path=dmriprep_path)
    npt.assert_equal(my_afq.data_frame.shape, (n_subjects * n_sessions, 11))


def test_AFQ_data():
    """
    Test with some actual data
    """
    tmpdir = nbtmp.InTemporaryDirectory()
    afd.organize_stanford_data(path=tmpdir.name)
    myafq = api.AFQ(dmriprep_path=op.join(tmpdir.name, 'stanford_hardi',
                                          'derivatives', 'dmriprep'),
                    sub_prefix='sub')
    npt.assert_equal(nib.load(myafq.b0[0]).shape,
                     nib.load(myafq['dwi_file'][0]).shape[:3])
    npt.assert_equal(nib.load(myafq.b0[0]).shape,
                     nib.load(myafq.dti[0]).shape[:3])


def test_AFQ_data_waypoint():
    """
    Test with some actual data again, this time for track segmentation
    """
    tmpdir = nbtmp.InTemporaryDirectory()
    afd.organize_stanford_data(path=tmpdir.name)
    dmriprep_path = op.join(tmpdir.name, 'stanford_hardi',
                            'derivatives', 'dmriprep')
    bundle_names = ["SLF", "ARC", "CST", "FP"]
    tracking_params = dict(odf_model="DTI")
    segmentation_params = dict(filter_by_endpoints=False,
                               seg_algo = "AFQ",
                               return_idx=True)

    clean_params = dict(return_idx=True)

    myafq = api.AFQ(dmriprep_path=dmriprep_path,
                    sub_prefix='sub',
                    bundle_names=bundle_names,
                    tracking_params=tracking_params,
                    segmentation_params=segmentation_params,
                    clean_params=clean_params)

    # Replace the mapping and streamlines with precomputed:
    file_dict = afd.read_stanford_hardi_tractography()
    mapping = file_dict['mapping.nii.gz']
    streamlines = file_dict['tractography_subsampled.trk']
    streamlines = dts.Streamlines(
        dtu.transform_tracking_output(
            [s for s in streamlines if s.shape[0] > 100],
            np.linalg.inv(myafq.dwi_affine[0])))

    sl_file = op.join(
        myafq.data_frame.results_dir[0],
        'sub-01_sess-01_dwi_space-RASMM_model-DTI_desc-det_tractography.trk')
    sft = StatefulTractogram(streamlines, myafq.data_frame.dwi_file[0],
                             Space.VOX)
    save_tractogram(sft, sl_file, bbox_valid_check=False)

    mapping_file = op.join(
        myafq.data_frame.results_dir[0],
        'sub-01_sess-01_dwi_mapping_from-DWI_to_MNI_xfm.nii.gz')
    nib.save(mapping, mapping_file)
    reg_prealign_file = op.join(
        myafq.data_frame.results_dir[0],
        'sub-01_sess-01_dwi_prealign_from-DWI_to-MNI_xfm.npy')
    np.save(reg_prealign_file, np.eye(4))

    tgram = load_tractogram(myafq.bundles[0], myafq.dwi_img[0])

    bundles = aus.tgram_to_bundles(tgram, myafq.bundle_dict, myafq.dwi_img[0])
    npt.assert_(len(bundles['CST_L']) > 0)

    # Test ROI exporting:
    myafq.export_rois()
    assert op.exists(op.join(
        myafq.data_frame['results_dir'][0],
        'ROIs',
        'sub-01_sess-01_dwi_desc-ROI-CST_R-1-include.json'))

    # Test bundles exporting:
    myafq.export_bundles()
    assert op.exists(op.join(
        myafq.data_frame['results_dir'][0],
        'bundles',
        'sub-01_sess-01_dwi_space-RASMM_model-DTI_desc-det-AFQ-CST_L_tractography.trk'))  # noqa

    # Test creation of file with bundle indices:
    assert op.exists(op.join(
        myafq.data_frame['results_dir'][0],
        'sub-01_sess-01_dwi_space-RASMM_model-DTI_desc-det-AFQ-clean_tractography_idx.json'))  # noqa

    tract_profiles = pd.read_csv(myafq.tract_profiles[0])
    assert tract_profiles.shape == (800, 5)

    # Before we run the CLI, we'll remove the bundles and ROI folders, to see
    # that the CLI generates them
    shutil.rmtree(op.join(myafq.data_frame['results_dir'][0],
                          'bundles'))

    shutil.rmtree(op.join(myafq.data_frame['results_dir'][0],
                          'ROIs'))

    # Test the CLI:
    print("Running the CLI:")

    # Bare bones config only points to the files
    config = dict(files=dict(dmriprep_path=dmriprep_path))

    config_file = op.join(tmpdir.name, "afq_config.toml")
    with open(config_file, 'w') as ff:
        toml.dump(config, ff)

    cmd = "pyAFQ " + config_file
    out = os.system(cmd)
    assert out == 0
    # The combined tract profiles should already exist from the CLI Run:
    from_file = pd.read_csv(op.join(myafq.afq_dir, 'tract_profiles.csv'))
    # And should be identical to what we would get by rerunning this:
    combined_profiles = myafq.combine_profiles()
    assert combined_profiles.shape == (800, 7)
    assert_frame_equal(combined_profiles, from_file)

    # Make sure the CLI did indeed generate these:
    myafq.export_rois()
    assert op.exists(op.join(
        myafq.data_frame['results_dir'][0],
        'ROIs',
        'sub-01_sess-01_dwi_desc-ROI-CST_R-1-include.json'))

    myafq.export_bundles()
    assert op.exists(op.join(
        myafq.data_frame['results_dir'][0],
        'bundles',
        'sub-01_sess-01_dwi_space-RASMM_model-DTI_desc-det-AFQ-CST_L_tractography.trk'))  # noqa


# def test_AFQ_data_recobundles():
#     tmpdir = nbtmp.InTemporaryDirectory()
#     afd.fetch_hcp(["100206"], hcp_bucket='hcp-openaccess', profile_name="hcp",
#                   path=tmpdir.name)
#     dmriprep_path = op.join(tmpdir.name, 'HCP', 'derivatives', 'dmriprep')
#     seg_algo = "reco"
#     bundle_names = ["F", "CST", "AF", "CC_ForcepsMajor"]
#     myafq = api.AFQ(dmriprep_path=dmriprep_path,
#                     sub_prefix='sub',
#                     seg_algo=seg_algo,
#                     bundle_names=bundle_names,
#                     odf_model="DTI",
#                     b0_threshold=15)

#     # Replace the streamlines with precomputed:
#     path_to_trk = dpd.fetcher.fetch_target_tractogram_hcp()
#     path_to_trk = dpd.fetcher.get_target_tractogram_hcp()
#     sl_file = op.join(myafq.data_frame.results_dir[0], 'sub-100206_sess-01_dwiDTI_det_streamlines.trk')
#     shutil.copy(path_to_trk, sl_file)
#     myafq.data_frame["streamlines_file"] = sl_file
#     print("here")
#     tgram = nib.streamlines.load(myafq.bundles[0]).tractogram
#     print("here")
#     bundles = aus.tgram_to_bundles(tgram, myafq.bundle_dict)
#     npt.assert_(len(bundles['CST_L']) > 0)



import os.path as op
import numpy as np
import numpy.testing as npt

import nibabel.tmpdirs as nbtmp

from AFQ.csd import fit_csd
from AFQ.dti import fit_dti
from AFQ.tractography import track
from AFQ.utils.testing import make_tracking_data


seeds = np.array([[-80., -120., -60.],
                  [-81, -121, -61],
                  [-81, -120, -60]])


tmpdir = nbtmp.InTemporaryDirectory()
fbval = op.join(tmpdir.name, 'dti.bval')
fbvec = op.join(tmpdir.name, 'dti.bvec')
fdata = op.join(tmpdir.name, 'dti.nii.gz')
make_tracking_data(fbval, fbvec, fdata)

min_length = 20
step_size = 0.5


def test_csd_tracking():
    for sh_order in [4, 8, 10]:
        fname = fit_csd(fdata, fbval, fbvec,
                        response=((0.0015, 0.0003, 0.0003), 100),
                        sh_order=8, lambda_=1, tau=0.1, mask=None,
                        out_dir=tmpdir.name)
        for directions in ["det", "prob"]:
            sl = track(fname, directions,
                       odf_model="CSD",
                       max_angle=30.,
                       sphere=None,
                       seed_mask=None,
                       n_seeds=seeds,
                       stop_mask=None,
                       step_size=step_size,
                       min_length=min_length).streamlines

            npt.assert_(len(sl[0]) >= step_size * min_length)


def test_dti_tracking():
    fdict = fit_dti(fdata, fbval, fbvec)
    for directions in ["det", "prob"]:
        sl = track(fdict['params'],
                   directions,
                   max_angle=30.,
                   sphere=None,
                   seed_mask=None,
                   n_seeds=1,
                   step_size=step_size,
                   min_length=min_length).streamlines
        npt.assert_(len(sl[0]) >= min_length * step_size)



import numpy as np
import multiprocessing
import joblib
import dask
import dask.multiprocessing


def parfor(func, in_list, out_shape=None, n_jobs=-1, engine="joblib",
           backend="threading", func_args=[], func_kwargs={}):
    """
    Parallel for loop for numpy arrays

    Parameters
    ----------
    func : callable
        The function to apply to each item in the array. Must have the form:
        func(arr, idx, *args, *kwargs) where arr is an ndarray and idx is an
        index into that array (a tuple). The Return of `func` needs to be one
        item (e.g. float, int) per input item.
    in_list : list
       All legitimate inputs to the function to operate over.
    n_jobs : integer, optional
        The number of jobs to perform in parallel. -1 to use all cpus.
        Default: 1
    engine : str
        {"dask", "joblib", "serial"}
        The last one is useful for debugging -- runs the code without any
        parallelization.
    backend : str
        What joblib backend to use. Irrelevant for other engines.
    func_args : list, optional
        Positional arguments to `func`.
    func_kwargs : list, optional
        Keyword arguments to `func`.

    Returns
    -------
    ndarray of identical shape to `arr`

    Examples
    --------

    """
    if n_jobs == -1:
        n_jobs = multiprocessing.cpu_count()
        n_jobs = n_jobs - 1

    if engine == "joblib":
        p = joblib.Parallel(n_jobs=n_jobs, backend=backend)
        d = joblib.delayed(func)
        d_l = []
        for in_element in in_list:
            d_l.append(d(in_element, *func_args, **func_kwargs))
        results = p(d_l)

    elif engine == "dask":
        def partial(func, *args, **keywords):
            def newfunc(in_arg):
                return func(in_arg, *args, **keywords)
            return newfunc
        p = partial(func, *func_args, **func_kwargs)
        d = [dask.delayed(p)(i) for i in in_list]
        if backend == "multiprocessing":
            results = dask.compute(*d, scheduler="processes",
                                   workers=n_jobs)
        elif backend == "threading":
            results = dask.compute(*d, scheduler="threads",
                                   workers=n_jobs)
        else:
            raise ValueError("%s is not a backend for dask" % backend)

    elif engine == "serial":
        results = []
        for in_element in in_list:
            results.append(func(in_element, *func_args, **func_kwargs))

    if out_shape is not None:
        return np.array(results).reshape(out_shape)
    else:
        return results



from setuptools import find_packages
import string
import os.path as op

try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

here = op.abspath(op.dirname(__file__))

# Get metadata from the AFQ/version.py file:
meta_file = op.join(here, 'AFQ', '_meta.py')
with open(meta_file) as f:
    exec(f.read())

REQUIRES = []
with open(op.join(here, 'requirements.txt')) as f:
    ll = f.readline()[:-1]
    while ll:
        REQUIRES.append(ll)
        ll = f.readline()[:-1]

with open(op.join(here, 'README.md'), encoding='utf-8') as f:
    LONG_DESCRIPTION = f.read()


def local_version(version):
    """
    Patch in a version that can be uploaded to test PyPI
    """
    gh_in_int = []
    for char in version.node:
        if char.isdigit():
            gh_in_int.append(str(char))
        else:
            gh_in_int.append(str(string.ascii_letters.find(char)))
    return "".join(gh_in_int)

opts = dict(name=NAME,
            maintainer=MAINTAINER,
            maintainer_email=MAINTAINER_EMAIL,
            description=DESCRIPTION,
            long_description=LONG_DESCRIPTION,
            url=URL,
            download_url=DOWNLOAD_URL,
            license=LICENSE,
            classifiers=CLASSIFIERS,
            author=AUTHOR,
            author_email=AUTHOR_EMAIL,
            platforms=PLATFORMS,
            packages=find_packages(),
            install_requires=REQUIRES,
            scripts=SCRIPTS,
            python_requires=PYTHON_REQUIRES,
            use_scm_version={"root": ".", "relative_to": __file__,
                             "write_to": "AFQ/version.py",
                             "local_scheme": local_version},
            setup_requires=['setuptools_scm'])


if __name__ == '__main__':
    setup(**opts)



import scipy.ndimage as ndim
from skimage.filters import gaussian
from skimage.morphology import convex_hull_image
from scipy.spatial.qhull import QhullError


def patch_up_roi(roi):
    """
    After being non-linearly transformed, ROIs tend to have holes in them.
    We perform a couple of computational geometry operations on the ROI to
    fix that up.

    Parameters
    ----------
    roi : 3D binary array
        The ROI after it has been transformed.

    sigma : float
        The sigma for initial Gaussian smoothing.

    truncate : float
        The truncation for the Gaussian

    Returns
    -------
    ROI after dilation and hole-filling
    """

    hole_filled = ndim.binary_fill_holes(roi > 0)
    try:
        return convex_hull_image(hole_filled)
    except QhullError:
        return hole_filled



"""
=========================================
Plotting tract profiles using RecoBundles
=========================================

An example of tracking and segmenting two tracts with RecoBundles
[Garyfallidis2017]_, and plotting their tract profiles for FA (calculated with
DTI).

See `plot_tract_profile` for explanations of each stage here. The main
difference here is that segmentation uses the RecoBundles algorithm, instead of
the AFQ waypoint ROI approach.

"""

import os.path as op
import matplotlib.pyplot as plt
import numpy as np
import nibabel as nib
import dipy.data as dpd
from dipy.data import fetcher
import dipy.tracking.utils as dtu
import dipy.tracking.streamline as dts
from dipy.io.streamline import save_tractogram, load_tractogram
from dipy.stats.analysis import afq_profile, gaussian_weights
from dipy.io.stateful_tractogram import StatefulTractogram
from dipy.io.stateful_tractogram import Space


import AFQ.data as afd
import AFQ.tractography as aft
import AFQ.registration as reg
import AFQ.dti as dti
import AFQ.segmentation as seg
import AFQ.api as api

dpd.fetch_stanford_hardi()

hardi_dir = op.join(fetcher.dipy_home, "stanford_hardi")
hardi_fdata = op.join(hardi_dir, "HARDI150.nii.gz")
hardi_fbval = op.join(hardi_dir, "HARDI150.bval")
hardi_fbvec = op.join(hardi_dir, "HARDI150.bvec")

img = nib.load(hardi_fdata)

print("Calculating DTI...")
if not op.exists('./dti_FA.nii.gz'):
    dti_params = dti.fit_dti(hardi_fdata, hardi_fbval, hardi_fbvec,
                             out_dir='.')
else:
    dti_params = {'FA': './dti_FA.nii.gz',
                  'params': './dti_params.nii.gz'}

FA_img = nib.load(dti_params['FA'])
FA_data = FA_img.get_fdata()

print("Registering to template...")
MNI_T2_img = dpd.read_mni_template()
if not op.exists('mapping.nii.gz'):
    import dipy.core.gradients as dpg
    gtab = dpg.gradient_table(hardi_fbval, hardi_fbvec)
    warped_hardi, mapping = reg.syn_register_dwi(hardi_fdata, gtab,
                                                 template=MNI_T2_img)
    reg.write_mapping(mapping, './mapping.nii.gz')
else:
    mapping = reg.read_mapping('./mapping.nii.gz', img, MNI_T2_img)

bundle_names = ["CST", "UF", "CC_ForcepsMajor", "CC_ForcepsMinor"]
bundles = api.make_bundle_dict(bundle_names=bundle_names, seg_algo="reco")

print("Tracking...")
if not op.exists('dti_streamlines_reco.trk'):
    seed_roi = np.zeros(img.shape[:-1])
    for bundle in bundles:
        if bundle != 'whole_brain':
            sl_xform = dts.Streamlines(
                dtu.transform_tracking_output(bundles[bundle]['sl'],
                                            MNI_T2_img.affine))

            delta = dts.values_from_volume(mapping.backward,
                                            sl_xform, np.eye(4))
            sl_xform = [sum(d, s) for d, s in zip(delta, sl_xform)]

            sl_xform = dts.Streamlines(
                dtu.transform_tracking_output(sl_xform,
                np.linalg.inv(MNI_T2_img.affine)))

            sft = StatefulTractogram(sl_xform, img, Space.RASMM)
            save_tractogram(sft, f'./{bundle}_atlas.trk')

            sl_xform = dts.Streamlines(
                dtu.transform_tracking_output(sl_xform,
                                        np.linalg.inv(img.affine)))

            for sl in sl_xform:
                sl_as_idx = sl.astype(int)
                seed_roi[sl_as_idx[:, 0],
                         sl_as_idx[:, 1],
                         sl_as_idx[:, 2]] = 1

    nib.save(nib.Nifti1Image(seed_roi, img.affine), 'seed_roi.nii.gz')
    sft = aft.track(dti_params['params'], seed_mask=seed_roi,
                    directions='det', stop_mask=FA_data,
                    stop_threshold=0.1)
    print(len(sft.streamlines))
    save_tractogram(sft, './dti_streamlines_reco.trk',
                    bbox_valid_check=False)
else:
    sft = load_tractogram('./dti_streamlines_reco.trk', img)

print("Segmenting fiber groups...")
segmentation = seg.Segmentation(seg_algo='reco',
                                rng=np.random.RandomState(2),
                                greater_than=50,
                                rm_small_clusters=10,
                                model_clust_thr=5,
                                reduction_thr=20,
                                refine=True)

segmentation.segment(bundles,
                     sft,
                     fdata=hardi_fdata,
                     fbval=hardi_fbval,
                     fbvec=hardi_fbvec,
                     mapping=mapping,
                     reg_template=MNI_T2_img)

fiber_groups = segmentation.fiber_groups

for kk in fiber_groups:
    print(kk, len(fiber_groups[kk]))
    sft = StatefulTractogram(fiber_groups[kk].streamlines,
                             img,
                             Space.RASMM)
    save_tractogram(sft, './%s_reco.trk'%kk,
                    bbox_valid_check=False)


print("Extracting tract profiles...")
for bundle in bundles:
    if bundle != 'whole_brain':

        fig, ax = plt.subplots(1)
        sft = load_tractogram(f'./{bundle}_reco.trk', img, to_space=Space.VOX)
        weights = gaussian_weights(sft.streamlines)
        profile = afq_profile(FA_data, sft.streamlines,
                              np.eye(4), weights=weights)

        ax.plot(profile)
        ax.set_title(bundle)

plt.show()

##########################################################################
# References:
# -------------------------
#  .. [Garyfallidis2017] Garyfallidis, Eleftherios, Marc-Alexandre Côté,
#                      Francois Rheault, Jasmeen Sidhu, Janice Hau, Laurent
#                      Petit, David Fortin, Stephen Cunanne, and Maxime
#                      Descoteaux. 2017.“Recognition of White Matter Bundles
#                      Using Local and Global Streamline-Based Registration and
#                      Clustering.”NeuroImage 170: 283-295.



import tempfile
import os.path as op
import numpy as np
import IPython.display as display

import nibabel as nib
from dipy.viz.colormap import line_colors
from dipy.viz import window, actor, ui

from palettable.tableau import Tableau_20

import AFQ.utils.volume as auv
import AFQ.registration as reg


def _inline_interact(ren, inline, interact):
    """
    Helper function to reuse across viz functions
    """
    if interact:
        window.show(ren)

    if inline:
        tdir = tempfile.gettempdir()
        fname = op.join(tdir, "fig.png")
        window.record(ren, out_path=fname, size=(1200, 1200))
        display.display_png(display.Image(fname))

    return ren


def visualize_bundles(trk, affine_or_mapping=None, bundle=None, ren=None,
                      color=None, inline=True, interact=False):
    """
    Visualize bundles in 3D using VTK
    """
    if isinstance(trk, str):
        trk = nib.streamlines.load(trk)
        tg = trk.tractogram
    else:
        # Assume these are streamlines (as list or Streamlines object):
        tg = nib.streamlines.Tractogram(trk)

    if affine_or_mapping is not None:
        tg = tg.apply_affine(np.linalg.inv(affine_or_mapping))

    streamlines = tg.streamlines

    if ren is None:
        ren = window.Renderer()

    # There are no bundles in here:
    if list(tg.data_per_streamline.keys()) == []:
        streamlines = list(streamlines)
        sl_actor = actor.line(streamlines, line_colors(streamlines))
        ren.add(sl_actor)
        sl_actor.GetProperty().SetRenderLinesAsTubes(1)
        sl_actor.GetProperty().SetLineWidth(6)
    if bundle is None:
        for b in np.unique(tg.data_per_streamline['bundle']):
            idx = np.where(tg.data_per_streamline['bundle'] == b)[0]
            this_sl = list(streamlines[idx])
            if color is not None:
                sl_actor = actor.line(this_sl, color)
                sl_actor.GetProperty().SetRenderLinesAsTubes(1)
                sl_actor.GetProperty().SetLineWidth(6)
            else:
                sl_actor = actor.line(this_sl,
                                      Tableau_20.colors[np.mod(20, int(b))])
                sl_actor.GetProperty().SetRenderLinesAsTubes(1)
                sl_actor.GetProperty().SetLineWidth(6)

            ren.add(sl_actor)
    else:
        idx = np.where(tg.data_per_streamline['bundle'] == bundle)[0]
        this_sl = list(streamlines[idx])
        if color is not None:
            sl_actor = actor.line(this_sl, color)
            sl_actor.GetProperty().SetRenderLinesAsTubes(1)
            sl_actor.GetProperty().SetLineWidth(6)

        else:
            sl_actor = actor.line(
                this_sl,
                Tableau_20.colors[np.mod(20, int(bundle))])
            sl_actor.GetProperty().SetRenderLinesAsTubes(1)
            sl_actor.GetProperty().SetLineWidth(6)
        ren.add(sl_actor)

    return _inline_interact(ren, inline, interact)


def visualize_roi(roi, affine_or_mapping=None, static_img=None,
                  roi_affine=None, static_affine=None, reg_template=None,
                  ren=None, color=None, inline=True, interact=False):
    """
    Render a region of interest into a VTK viz as a volume
    """
    if not isinstance(roi, np.ndarray):
        if isinstance(roi, str):
            roi = nib.load(roi).get_fdata()
        else:
            roi = roi.get_fdata()

    if affine_or_mapping is not None:
        if isinstance(affine_or_mapping, np.ndarray):
            # This is an affine:
            if (static_img is None or roi_affine is None
                    or static_affine is None):
                raise ValueError("If using an affine to transform an ROI, "
                                 "need to also specify all of the following",
                                 "inputs: `static_img`, `roi_affine`, ",
                                 "`static_affine`")
            roi = reg.resample(roi, static_img, roi_affine, static_affine)
        else:
            # Assume it is  a mapping:
            if (isinstance(affine_or_mapping, str)
                    or isinstance(affine_or_mapping, nib.Nifti1Image)):
                if reg_template is None or static_img is None:
                    raise ValueError(
                        "If using a mapping to transform an ROI, need to ",
                        "also specify all of the following inputs: ",
                        "`reg_template`, `static_img`")
                affine_or_mapping = reg.read_mapping(affine_or_mapping,
                                                     static_img,
                                                     reg_template)

            roi = auv.patch_up_roi(affine_or_mapping.transform_inverse(
                                   roi,
                                   interpolation='nearest')).astype(bool)

    if ren is None:
        ren = window.Renderer()

    roi_actor = actor.contour_from_roi(roi, color=color)
    ren.add(roi_actor)

    if inline:
        tdir = tempfile.gettempdir()
        fname = op.join(tdir, "fig.png")
        window.record(ren, out_path=fname)
        display.display_png(display.Image(fname))

    return _inline_interact(ren, inline, interact)


def visualize_volume(volume, x=None, y=None, z=None, ren=None, inline=True,
                     interact=False):
    """
    Visualize a volume
    """
    if ren is None:
        ren = window.Renderer()

    shape = volume.shape
    image_actor_z = actor.slicer(volume)
    slicer_opacity = 0.6
    image_actor_z.opacity(slicer_opacity)

    image_actor_x = image_actor_z.copy()
    x_midpoint = int(np.round(shape[0] / 2))
    image_actor_x.display_extent(x_midpoint,
                                 x_midpoint, 0,
                                 shape[1] - 1,
                                 0,
                                 shape[2] - 1)

    image_actor_y = image_actor_z.copy()
    y_midpoint = int(np.round(shape[1] / 2))
    image_actor_y.display_extent(0,
                                 shape[0] - 1,
                                 y_midpoint,
                                 y_midpoint,
                                 0,
                                 shape[2] - 1)

    ren.add(image_actor_z)
    ren.add(image_actor_x)
    ren.add(image_actor_y)

    show_m = window.ShowManager(ren, size=(1200, 900))
    show_m.initialize()

    line_slider_z = ui.LineSlider2D(min_value=0,
                                    max_value=shape[2] - 1,
                                    initial_value=shape[2] / 2,
                                    text_template="{value:.0f}",
                                    length=140)

    line_slider_x = ui.LineSlider2D(min_value=0,
                                    max_value=shape[0] - 1,
                                    initial_value=shape[0] / 2,
                                    text_template="{value:.0f}",
                                    length=140)

    line_slider_y = ui.LineSlider2D(min_value=0,
                                    max_value=shape[1] - 1,
                                    initial_value=shape[1] / 2,
                                    text_template="{value:.0f}",
                                    length=140)

    opacity_slider = ui.LineSlider2D(min_value=0.0,
                                     max_value=1.0,
                                     initial_value=slicer_opacity,
                                     length=140)

    def change_slice_z(slider):
        z = int(np.round(slider.value))
        image_actor_z.display_extent(0, shape[0] - 1, 0, shape[1] - 1, z, z)

    def change_slice_x(slider):
        x = int(np.round(slider.value))
        image_actor_x.display_extent(x, x, 0, shape[1] - 1, 0, shape[2] - 1)

    def change_slice_y(slider):
        y = int(np.round(slider.value))
        image_actor_y.display_extent(0, shape[0] - 1, y, y, 0, shape[2] - 1)

    def change_opacity(slider):
        slicer_opacity = slider.value
        image_actor_z.opacity(slicer_opacity)
        image_actor_x.opacity(slicer_opacity)
        image_actor_y.opacity(slicer_opacity)

    line_slider_z.on_change = change_slice_z
    line_slider_x.on_change = change_slice_x
    line_slider_y.on_change = change_slice_y
    opacity_slider.on_change = change_opacity

    def build_label(text):
        label = ui.TextBlock2D()
        label.message = text
        label.font_size = 18
        label.font_family = 'Arial'
        label.justification = 'left'
        label.bold = False
        label.italic = False
        label.shadow = False
        label.background = (0, 0, 0)
        label.color = (1, 1, 1)

        return label

    line_slider_label_z = build_label(text="Z Slice")
    line_slider_label_x = build_label(text="X Slice")
    line_slider_label_y = build_label(text="Y Slice")
    opacity_slider_label = build_label(text="Opacity")

    panel = ui.Panel2D(size=(300, 200),
                       color=(1, 1, 1),
                       opacity=0.1,
                       align="right")
    panel.center = (1030, 120)

    panel.add_element(line_slider_label_x, (0.1, 0.75))
    panel.add_element(line_slider_x, (0.38, 0.75))
    panel.add_element(line_slider_label_y, (0.1, 0.55))
    panel.add_element(line_slider_y, (0.38, 0.55))
    panel.add_element(line_slider_label_z, (0.1, 0.35))
    panel.add_element(line_slider_z, (0.38, 0.35))
    panel.add_element(opacity_slider_label, (0.1, 0.15))
    panel.add_element(opacity_slider, (0.38, 0.15))

    show_m.ren.add(panel)

    global size
    size = ren.GetSize()

    def win_callback(obj, event):
        global size
        if size != obj.GetSize():
            size_old = size
            size = obj.GetSize()
            size_change = [size[0] - size_old[0], 0]
            panel.re_align(size_change)

    show_m.initialize()

    ren.zoom(1.5)
    ren.reset_clipping_range()

    if interact:
        show_m.add_window_callback(win_callback)
        show_m.render()
        show_m.start()

    return _inline_interact(ren, inline, interact)



"""
Registration tools
"""
import os
import os.path as op
import numpy as np
import nibabel as nib
from dipy.align.metrics import CCMetric, EMMetric, SSDMetric
from dipy.align.imwarp import (SymmetricDiffeomorphicRegistration,
                               DiffeomorphicMap)

from dipy.align.imaffine import (transform_centers_of_mass,
                                 AffineMap,
                                 MutualInformationMetric,
                                 AffineRegistration)

from dipy.align.transforms import (TranslationTransform3D,
                                   RigidTransform3D,
                                   AffineTransform3D)

import dipy.core.gradients as dpg
import dipy.data as dpd
from dipy.align.streamlinear import StreamlineLinearRegistration
from dipy.tracking.streamline import set_number_of_points
from dipy.tracking.utils import transform_tracking_output
from dipy.io.streamline import load_tractogram, load_trk

import AFQ.utils.models as mut
import AFQ.utils.streamlines as sut

syn_metric_dict = {'CC': CCMetric,
                   'EM': EMMetric,
                   'SSD': SSDMetric}

__all__ = ["syn_registration", "syn_register_dwi", "write_mapping",
           "read_mapping", "resample", "c_of_mass", "translation", "rigid",
           "affine", "affine_registration", "register_series", "register_dwi",
           "streamline_registration"]


def syn_registration(moving, static,
                     moving_affine=None,
                     static_affine=None,
                     step_length=0.25,
                     metric='CC',
                     dim=3,
                     level_iters=[10, 10, 5],
                     sigma_diff=2.0,
                     radius=4,
                     prealign=None):
    """Register a source image (moving) to a target image (static).

    Parameters
    ----------
    moving : ndarray
        The source image data to be registered
    moving_affine : array, shape (4,4)
        The affine matrix associated with the moving (source) data.
    static : ndarray
        The target image data for registration
    static_affine : array, shape (4,4)
        The affine matrix associated with the static (target) data
    metric : string, optional
        The metric to be optimized. One of `CC`, `EM`, `SSD`,
        Default: CCMetric.
    dim: int (either 2 or 3), optional
       The dimensions of the image domain. Default: 3
    level_iters : list of int, optional
        the number of iterations at each level of the Gaussian Pyramid (the
        length of the list defines the number of pyramid levels to be
        used).
    sigma_diff, radius : float
        Parameters for initialization of the metric.

    Returns
    -------
    warped_moving : ndarray
        The data in `moving`, warped towards the `static` data.
    forward : ndarray (..., 3)
        The vector field describing the forward warping from the source to the
        target.
    backward : ndarray (..., 3)
        The vector field describing the backward warping from the target to the
        source.
    """
    use_metric = syn_metric_dict[metric](dim, sigma_diff=sigma_diff,
                                         radius=radius)

    sdr = SymmetricDiffeomorphicRegistration(use_metric, level_iters,
                                             step_length=step_length)

    mapping = sdr.optimize(static, moving,
                           static_grid2world=static_affine,
                           moving_grid2world=moving_affine,
                           prealign=prealign)

    warped_moving = mapping.transform(moving)
    return warped_moving, mapping


def syn_register_dwi(dwi, gtab, template=None, **syn_kwargs):
    """
    Register DWI data to a template.

    Parameters
    -----------
    dwi : nifti image or str
        Image containing DWI data, or full path to a nifti file with DWI.
    gtab : GradientTable or list of strings
        The gradients associated with the DWI data, or a string with [fbcal, ]
    template : nifti image or str, optional

    syn_kwargs : key-word arguments for :func:`syn_registration`

    Returns
    -------
    DiffeomorphicMap object
    """
    if template is None:
        template = dpd.read_mni_template()
    if isinstance(template, str):
        template = nib.load(template)

    template_data = template.get_fdata()
    template_affine = template.affine

    if isinstance(dwi, str):
        dwi = nib.load(dwi)

    if not isinstance(gtab, dpg.GradientTable):
        gtab = dpg.gradient_table(*gtab)

    dwi_affine = dwi.affine
    dwi_data = dwi.get_fdata()
    mean_b0 = np.mean(dwi_data[..., gtab.b0s_mask], -1)
    warped_b0, mapping = syn_registration(mean_b0, template_data,
                                          moving_affine=dwi_affine,
                                          static_affine=template_affine,
                                          **syn_kwargs)
    return warped_b0, mapping


def write_mapping(mapping, fname):
    """
    Write out a syn registration mapping to file

    Parameters
    ----------
    mapping : a DiffeomorphicMap object derived from :func:`syn_registration`
    fname : str
        Full path to the nifti file storing the mapping

    """
    mapping_data = np.array([mapping.forward.T, mapping.backward.T]).T
    nib.save(nib.Nifti1Image(mapping_data, mapping.codomain_world2grid), fname)


def read_mapping(disp, domain_img, codomain_img, prealign=None):
    """
    Read a syn registration mapping from a nifti file

    Parameters
    ----------
    disp : str or Nifti1Image
        A file of image containing the mapping displacement field in each voxel
        Shape (x, y, z, 3, 2)

    domain_img : str or Nifti1Image

    codomain_img : str or Nifti1Image

    Returns
    -------
    A :class:`DiffeomorphicMap` object
    """
    if isinstance(disp, str):
        disp = nib.load(disp)

    if isinstance(domain_img, str):
        domain_img = nib.load(domain_img)

    if isinstance(codomain_img, str):
        codomain_img = nib.load(codomain_img)

    mapping = DiffeomorphicMap(3, disp.shape[:3],
                               disp_grid2world=np.linalg.inv(disp.affine),
                               domain_shape=domain_img.shape[:3],
                               domain_grid2world=domain_img.affine,
                               codomain_shape=codomain_img.shape,
                               codomain_grid2world=codomain_img.affine,
                               prealign=prealign)

    disp_data = disp.get_fdata().astype(np.float32)
    mapping.forward = disp_data[..., 0]
    mapping.backward = disp_data[..., 1]
    mapping.is_inverse = True

    return mapping


def resample(moving, static, moving_affine, static_affine):
    """Resample an image from one space to another.

    Parameters
    ----------
    moving : array
       The image to be resampled

    static : array

    moving_affine
    static_affine

    Returns
    -------
    resampled : the moving array resampled into the static array's space.
    """
    identity = np.eye(4)
    affine_map = AffineMap(identity,
                           static.shape, static_affine,
                           moving.shape, moving_affine)
    resampled = affine_map.transform(moving)
    return resampled


# Affine registration pipeline:
affine_metric_dict = {'MI': MutualInformationMetric}


def c_of_mass(moving, static, static_affine, moving_affine,
              reg, starting_affine, params0=None):
    transform = transform_centers_of_mass(static, static_affine,
                                          moving, moving_affine)
    transformed = transform.transform(moving)
    return transformed, transform.affine


def translation(moving, static, static_affine, moving_affine,
                reg, starting_affine, params0=None):
    transform = TranslationTransform3D()
    translation = reg.optimize(static, moving, transform, params0,
                               static_affine, moving_affine,
                               starting_affine=starting_affine)

    return translation.transform(moving), translation.affine


def rigid(moving, static, static_affine, moving_affine,
          reg, starting_affine, params0=None):
    transform = RigidTransform3D()
    rigid = reg.optimize(static, moving, transform, params0,
                         static_affine, moving_affine,
                         starting_affine=starting_affine)
    return rigid.transform(moving), rigid.affine


def affine(moving, static, static_affine, moving_affine,
           reg, starting_affine, params0=None):
    transform = AffineTransform3D()
    affine = reg.optimize(static, moving, transform, params0,
                          static_affine, moving_affine,
                          starting_affine=starting_affine)

    return affine.transform(moving), affine.affine


def affine_registration(moving, static,
                        moving_affine=None,
                        static_affine=None,
                        nbins=32,
                        sampling_prop=None,
                        metric='MI',
                        pipeline=[c_of_mass, translation, rigid, affine],
                        level_iters=[10000, 1000, 100],
                        sigmas=[5.0, 2.5, 0.0],
                        factors=[4, 2, 1],
                        params0=None):

    """
    Find the affine transformation between two 3D images.

    Parameters
    ----------

    """
    # Define the Affine registration object we'll use with the chosen metric:
    use_metric = affine_metric_dict[metric](nbins, sampling_prop)
    affreg = AffineRegistration(metric=use_metric,
                                level_iters=level_iters,
                                sigmas=sigmas,
                                factors=factors)

    # Bootstrap this thing with the identity:
    starting_affine = np.eye(4)
    # Go through the selected transformation:
    for func in pipeline:
        transformed, starting_affine = func(moving, static,
                                            static_affine,
                                            moving_affine,
                                            affreg, starting_affine,
                                            params0)
    return transformed, starting_affine


def register_series(series, ref, pipeline):
    """Register a series to a reference image.

    Parameters
    ----------
    series : Nifti1Image object
       The data is 4D with the last dimension separating different 3D volumes
    ref : Nifti1Image or integer or iterable

    Returns
    -------
    transformed_list, affine_list
    """
    if isinstance(ref, nib.Nifti1Image):
        static = ref
        static_data = static.get_fdata()
        s_aff = static.affine
        moving = series
        moving_data = moving.get_fdata()
        m_aff = moving.affine

    elif isinstance(ref, int) or np.iterable(ref):
        data = series.get_fdata()
        idxer = np.zeros(data.shape[-1]).astype(bool)
        idxer[ref] = True
        static_data = data[..., idxer]
        if len(static_data.shape) > 3:
            static_data = np.mean(static_data, -1)
        moving_data = data[..., ~idxer]
        m_aff = s_aff = series.affine

    affine_list = []
    transformed_list = []
    for ii in range(moving_data.shape[-1]):
        this_moving = moving_data[..., ii]
        transformed, affine = affine_registration(this_moving, static_data,
                                                  moving_affine=m_aff,
                                                  static_affine=s_aff,
                                                  pipeline=pipeline)
        transformed_list.append(transformed)
        affine_list.append(affine)

    return transformed_list, affine_list


def register_dwi(data_files, bval_files, bvec_files,
                 b0_ref=0,
                 pipeline=[c_of_mass, translation, rigid, affine],
                 out_dir=None):
    """
    Register a DWI data-set

    Parameters
    ----------
    data_files : str or list
        Files containing DWI data. If this is a str, that's the full path to a
        single file. If it's a list, each entry is a full path.
    bval_files : str or list
        Equivalent to `data_files`.
    bvec_files : str or list
        Equivalent to `data_files`.


    """
    img, data, gtab, mask = mut.prepare_data(data_files,
                                             bval_files,
                                             bvec_files)
    if np.sum(gtab.b0s_mask) > 1:
        # First, register the b0s into one image:
        b0_img = nib.Nifti1Image(data[..., gtab.b0s_mask], img.affine)
        trans_b0 = register_series(b0_img, ref=0, pipeline=pipeline)
        ref_data = np.mean(trans_b0, -1)
    else:
        ref_data = data[..., gtab.b0s_mask]

    # Construct a series out of the DWI and the registered mean B0:
    series = nib.Nifti1Image(np.concatenate([ref_data,
                                             data[...,
                                                  ~gtab.b0s_mask]], -1),
                             img.affine)

    transformed_list, affine_list = register_series(series, ref=0,
                                                    pipeline=pipeline)
    reg_img = nib.Nifti1Image(np.array(transformed_list),
                              img.affine)

    if out_dir is None:
        out_dir = op.join(op.split(data_files)[0], 'registered')

    if not op.exists(out_dir):
        os.makedirs(out_dir)

    path = op.join(out_dir, 'registered.nii.gz')
    nib.save(reg_img, path)
    return path


def streamline_registration(moving, static, n_points=100,
                            native_resampled=False):
    """
    Register two collections of streamlines ('bundles') to each other

    Parameters
    ----------
    moving, static : lists of 3 by n, or str
        The two bundles to be registered. Given either as lists of arrays with
        3D coordinates, or strings containing full paths to these files.

    n_points : int, optional
        How many points to resample to. Default: 100.

    native_resampled : bool, optional
        Whether to return the moving bundle in the original space, but
        resampled in the static space to n_points.

    Returns
    -------
    aligned : list
        Streamlines from the moving group, moved to be closely matched to
        the static group.

    matrix : array (4, 4)
        The affine transformation that takes us from 'moving' to 'static'
    """
    # Load the streamlines, if you were given a file-name
    if isinstance(moving, str):
        moving = load_trk(moving, 'same', bbox_valid_check=False).streamlines
    if isinstance(static, str):
        static = load_trk(static, 'same', bbox_valid_check=False).streamlines

    srr = StreamlineLinearRegistration()
    srm = srr.optimize(static=set_number_of_points(static, n_points),
                       moving=set_number_of_points(moving, n_points))

    aligned = srm.transform(moving)
    if native_resampled:
        aligned = set_number_of_points(aligned, n_points)
        aligned = transform_tracking_output(aligned, np.linalg.inv(srm.matrix))

    return aligned, srm.matrix
